#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 Shreyas Kulkarni (shyran@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
from lstree.nodetree.builder import NodeTreeBuilder
from lstree.nodetree.plotter import NodeTreePlotter
from lstree.treeconfig import TreeConfig

__author__ = 'Shreyas Kulkarni'
__email__ = 'shyran@gmail.com'


def main():
    config = TreeConfig()
    plotter = NodeTreePlotter(config)

    # we can work with multiple root folders
    # (./ being the default if none is provided)
    for rf in config.folders():
        # a minor tweak to sweeten the output
        root_folder = rf.strip(os.path.sep)

        try:
            builder = NodeTreeBuilder(root_folder, config)
            root_node = builder.build()
            root_node.visit(plotter)

        except RuntimeError as rte:
            print("error: " + rte.message)
            return -1


if __name__ == '__main__':
    sys.exit(main() or 0)



# -*- coding: utf-8 -*-

# Copyright (c) 2016 Shreyas Kulkarni (shyran@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from argparse import ArgumentParser

__author__ = 'Shreyas Kulkarni'
__email__ = 'shyran@gmail.com'


class TreeConfig(object):
    def __init__(self):
        self._config = None
        self._parse_args()

    def _parse_args(self):
        parser = ArgumentParser()
        parser.add_argument('-s', '--show-hidden',
                            help="list hidden files and folders",
                            action='store_true',
                            default=False)

        parser.add_argument('--terse',
                            help="make it terse, visual pleasure is not desired",
                            action='store_true',
                            default=False)

        parser.add_argument('folders',
                            help='folders to draw tree for',
                            nargs='*',
                            default=['.'])

        parser.add_argument('-i', '--ignore',
                            help='ignore any file or folder that matches these wildcards',
                            nargs='*',
                            default=[])

        parser.add_argument('-f', '--filter',
                            help="filter and show *ONLY FILES* that match these wildcards",
                            nargs='+',
                            default=['*'])

        parser.add_argument('--ignore-empty',
                            help='ignore any empty folder (after filtering)',
                            action='store_true',
                            default=False)

        parser.add_argument('--tab',
                            help='how many spaces per tab. more the spaces, more spread out the tree',
                            type=int,
                            default=3)

        self._config = parser.parse_args()

    def show_hidden(self):
        return self._config.show_hidden

    def ignore_empty_directories(self):
        return self._config.ignore_empty

    def filter_patterns(self):
        return self._config.filter

    def ignore_patterns(self):
        return self._config.ignore

    def tabbing(self):
        return self._config.tab

    def folders(self):
        return self._config.folders

    def terse(self):
        return self._config.terse



# -*- coding: utf-8 -*-

# Copyright (c) 2016 Shreyas Kulkarni (shyran@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import abc

from lstree.nodetree.visitor import NodeTreeVisitor

__author__ = 'Shreyas Kulkarni'
__email__ = 'shyran@gmail.com'


class Node(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        self._name = name
        self._children = []

    def name(self):
        return self._name

    def __repr__(self):
        return self.name()

    def __add__(self, other):
        assert issubclass(other.__class__, Node)
        self._children.append(other)

        return self

    @abc.abstractmethod
    def empty(self):
        pass

    def prune(self, predicate):
        """
        allow a visitor predicate to choose candidates to prune
        then prune the chosen ones out
        :param predicate: any function that takes in a Node and returns True or False
        :return: nothing
        """
        # send it downwards first. process at the leaf nodes first, then bubble up
        for c in self._children:
            c.prune(predicate)

        # now actually prune children that match the predicate
        self._children = [c for c in self._children if not predicate(c)]

    def visit(self, visitor, topdown=True):
        """
        allow a visitor to visit each node in the node-tree in a top-down fashion (by default)
        :param visitor: an instance of NodeTreeVisitor which knows what to do when visiting each type of the nodes
        :param topdown: from top towards bottom if True, bottom-up if False
        :return: nothing
        """
        assert isinstance(visitor, NodeTreeVisitor)

        if topdown:
            # first visit self, if it's top-down
            visitor.visit(self)

        if len(self._children):
            visitor.down()

            # visit the children
            for c in self._children:
                c.visit(visitor, topdown)

            visitor.up()

        if not topdown:
            # if it's bottom-up, then visit self last
            visitor.visit(self)



# -*- coding: utf-8 -*-

# Copyright (c) 2016 Shreyas Kulkarni (shyran@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from fnmatch import fnmatch
import os

from lstree.nodes.directory import DirectoryNode
from lstree.nodes.file import FileNode
from lstree.treeconfig import TreeConfig

__author__ = 'Shreyas Kulkarni'
__email__ = 'shyran@gmail.com'


class NodeTreeBuilder(object):
    def __init__(self, root_dir, config):
        assert isinstance(config, TreeConfig)

        self._root_dir = root_dir
        self._config = config

        self._load_os_walk_cache()

    def _load_os_walk_cache(self):
        """
        use os.walk to get files and folders within the root dir
        builds a usable map before building a tree

        :return: nothing (changes state)
        """
        self._os_walk_cache = {a[0]: (a[1], a[2]) for a in os.walk(self._root_dir)}
        # print str(self._os_walk_cache)

    def build(self):
        """
        :param root_dir: root directory to start building from
        :return: built node tree
        """
        # build the node tree starting from the root directory
        root = self._build_directory(self._root_dir)

        # apply all filters
        if not self._config.show_hidden():
            root.prune(lambda node: node.name().startswith('.') and len(node.name()) > 0)

        # 'and'ing ignore patterns
        for pattern in self._config.ignore_patterns():
            root.prune(lambda node: fnmatch(node.name(), pattern))

        # 'or'ing the filter patterns
        root.prune(lambda node: isinstance(node, FileNode) and
                                not any([fnmatch(node.name(), pattern)
                                         for pattern in self._config.filter_patterns()]))

        if self._config.ignore_empty_directories():
            root.prune(lambda node: isinstance(node, DirectoryNode) and node.empty())

        return root

    def _build_directory(self, dirname):
        """
        :param dirname: directory name
        :param level: tabbing level for the directory
        :return: built DirectoryNode
        """
        # check if all's well
        if dirname not in self._os_walk_cache:
            raise RuntimeError("directory " + dirname + " not found")

        dnode = DirectoryNode(os.path.basename(dirname))

        dirs, files = self._os_walk_cache[dirname]
        for d in dirs:
            dnode += self._build_directory(os.path.join(dirname, d))

        for f in files:
            dnode += FileNode(f)

        return dnode



# -*- coding: utf-8 -*-

# Copyright (c) 2016 Shreyas Kulkarni (shyran@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase
import unittest
from lstree.nodetree.builder import NodeTreeBuilder
from lstree.nodetree.visitor import NodeTreeVisitor
from lstree.treeconfig import TreeConfig

__author__ = 'Shreyas Kulkarni'
__email__ = 'shyran@gmail.com'


class TestNodeTreeBuilder(TestCase):
    class PatchedNodeTreeBuilder(NodeTreeBuilder):
        def _load_os_walk_cache(self):
            self._os_walk_cache = {'./.hiddendir': ([],
                                                    ['.hiddenfile', '.hiddenfile.compiled']),
                                   './emptydir': ([], []),
                                   './somedir': ([],
                                                 ['somefile.compiled', 'somefile1', 'somefile2', 'somefile3']),
                                   '.': (['.hiddendir', 'emptydir', 'somedir'],
                                         ['datafile.xml', 'ignore.me.compiled', 'testfile'])}

    class MockedTreeConfig(TreeConfig):
        """
        mocking with unittest.mock is not useful since the builder checks instance of
        incoming config with isinstance. so we inherit to mock
        """
        def __init__(self,
                     show_hidden=False,
                     ignore_empty_directories=False,
                     filter_patterns=['*'],
                     ignore_patterns=[],
                     tabbing=3,
                     folders=["doesn't matter"],
                     terse=False,):
            self._show_hidden = show_hidden
            self._ignore_empty_directories = ignore_empty_directories
            self._filter_patterns = filter_patterns
            self._ignore_patterns = ignore_patterns
            self._tabbing = tabbing
            self._folders = folders
            self._terse = terse

        def terse(self): return self._terse

        def filter_patterns(self): return self._filter_patterns

        def folders(self): return self._folders

        def tabbing(self): return self._tabbing

        def ignore_empty_directories(self): return self._ignore_empty_directories

        def ignore_patterns(self): return self._ignore_patterns

        def show_hidden(self): return self._show_hidden

    class NodesAccumulator(NodeTreeVisitor):
        def __init__(self):
            # this is where we collect all nodes
            self._nodes_cache = set()

        def visit(self, node):
            super(TestNodeTreeBuilder.NodesAccumulator, self).visit(node)
            self._nodes_cache.add(node.name())

        def __contains__(self, item):
            return item in self._nodes_cache

    def test_build(self):
        mocked_config = self.MockedTreeConfig()
        patched_builder = self.PatchedNodeTreeBuilder(root_dir=".",
                                                      config=mocked_config)

        root = patched_builder.build()

        nodes = self.NodesAccumulator()
        root.visit(nodes)

        # ensure what's in
        self.assertTrue('emptydir' in nodes)
        self.assertTrue('somedir' in nodes)
        self.assertTrue('somefile.compiled' in nodes)
        self.assertTrue('somefile1' in nodes)
        self.assertTrue('somefile2' in nodes)
        self.assertTrue('somefile3' in nodes)
        self.assertTrue('datafile.xml' in nodes)
        self.assertTrue('ignore.me.compiled' in nodes)
        self.assertTrue('testfile' in nodes)

        # also ensure what stays out
        self.assertTrue('.hiddendir' not in nodes)
        self.assertTrue('.hiddenfile' not in nodes)
        self.assertTrue('.hiddenfile.compiled' not in nodes)

    def test_build_show_hidden(self):
        mocked_config = self.MockedTreeConfig(show_hidden=True)
        patched_builder = self.PatchedNodeTreeBuilder(root_dir=".",
                                                      config=mocked_config)

        root = patched_builder.build()

        nodes = self.NodesAccumulator()
        root.visit(nodes)

        # ensure what's in
        self.assertTrue('emptydir' in nodes)
        self.assertTrue('somedir' in nodes)
        self.assertTrue('somefile.compiled' in nodes)
        self.assertTrue('somefile1' in nodes)
        self.assertTrue('somefile2' in nodes)
        self.assertTrue('somefile3' in nodes)
        self.assertTrue('datafile.xml' in nodes)
        self.assertTrue('ignore.me.compiled' in nodes)
        self.assertTrue('testfile' in nodes)

        # also ensure that hidden files are listed as well
        self.assertTrue('.hiddendir' in nodes)
        self.assertTrue('.hiddenfile' in nodes)
        self.assertTrue('.hiddenfile.compiled' in nodes)

    def test_build_ignore_empty(self):
        mocked_config = self.MockedTreeConfig(ignore_empty_directories=True)
        patched_builder = self.PatchedNodeTreeBuilder(root_dir=".",
                                                      config=mocked_config)

        root = patched_builder.build()

        nodes = self.NodesAccumulator()
        root.visit(nodes)

        # ensure what's in
        self.assertTrue('somedir' in nodes)
        self.assertTrue('somefile.compiled' in nodes)
        self.assertTrue('somefile1' in nodes)
        self.assertTrue('somefile2' in nodes)
        self.assertTrue('somefile3' in nodes)
        self.assertTrue('datafile.xml' in nodes)
        self.assertTrue('ignore.me.compiled' in nodes)
        self.assertTrue('testfile' in nodes)

        # also ensure that hidden files are listed as well
        self.assertTrue('emptydir' not in nodes)
        self.assertTrue('.hiddendir' not in nodes)
        self.assertTrue('.hiddenfile' not in nodes)
        self.assertTrue('.hiddenfile.compiled' not in nodes)

    def test_build_ignore_patterns(self):
        """
        we can supply multiple wildcard patterns to be ignored
        they can be: file.*, partialname*, directoryname etc
        """
        mocked_config = self.MockedTreeConfig(ignore_patterns=['*.compiled',
                                                               'data*',
                                                               'somedir'])
        patched_builder = self.PatchedNodeTreeBuilder(root_dir=".",
                                                      config=mocked_config)

        root = patched_builder.build()

        nodes = self.NodesAccumulator()
        root.visit(nodes)

        # ensure what's in
        self.assertTrue('emptydir' in nodes)
        self.assertTrue('testfile' in nodes)

        # also ensure that hidden files are listed as well
        self.assertTrue('somedir' not in nodes)
        self.assertTrue('somefile1' not in nodes)
        self.assertTrue('somefile2' not in nodes)
        self.assertTrue('somefile3' not in nodes)
        self.assertTrue('datafile.xml' not in nodes)
        self.assertTrue('somefile.compiled' not in nodes)
        self.assertTrue('ignore.me.compiled' not in nodes)
        self.assertTrue('.hiddendir' not in nodes)
        self.assertTrue('.hiddenfile' not in nodes)
        self.assertTrue('.hiddenfile.compiled' not in nodes)

    def test_build_filter_patterns(self):
        """
        filter patterns are used as positive filters, i.e. to seek out
        only those files that match the given pattern
        these patterns can be specified similar to ignore patterns
        """
        mocked_config = self.MockedTreeConfig(filter_patterns=['*.compiled',
                                                               'data*'])
        patched_builder = self.PatchedNodeTreeBuilder(root_dir=".",
                                                      config=mocked_config)

        root = patched_builder.build()

        nodes = self.NodesAccumulator()
        root.visit(nodes)

        # ensure what's in
        self.assertTrue('emptydir' in nodes)        # empty dirs are still listed (unless ignored)
        self.assertTrue('somedir' in nodes)

        self.assertTrue('datafile.xml' in nodes)
        self.assertTrue('somefile.compiled' in nodes)
        self.assertTrue('ignore.me.compiled' in nodes)

        # also ensure that hidden files are listed as well
        self.assertTrue('testfile' not in nodes)
        self.assertTrue('somefile1' not in nodes)
        self.assertTrue('somefile2' not in nodes)
        self.assertTrue('somefile3' not in nodes)
        self.assertTrue('.hiddendir' not in nodes)
        self.assertTrue('.hiddenfile' not in nodes)
        self.assertTrue('.hiddenfile.compiled' not in nodes)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())
